@@ -33,6 +33,8 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 
 #include <QDebug>
 
+#include <functional>
+
 namespace KWayland
 {
 namespace Tools
@@ -172,16 +174,6 @@ Request::Request(const QString &name)
 
 Request::~Request() = default;
 
-bool Request::isDestructor() const
-{
-    for (const auto a: m_arguments) {
-        if (a.type() == Argument::Type::Destructor) {
-            return true;
-        }
-    }
-    return false;
-}
-
 bool Request::isFactory() const
 {
     for (const auto a: m_arguments) {
@@ -208,6 +200,7 @@ Interface::Interface() = default;
 Interface::Interface(const QXmlStreamAttributes &attributes)
     : m_name(attributes.value(QStringLiteral("name")).toString())
     , m_version(attributes.value(QStringLiteral("version")).toUInt())
+    , m_factory(Q_NULLPTR)
 {
     auto it = s_clientClassNameMapping.constFind(m_name);
     if (it != s_clientClassNameMapping.constEnd()) {
@@ -329,6 +322,9 @@ Request Generator::parseRequest()
 {
     const auto attributes = m_xmlReader.attributes();
     Request request(attributes.value(QStringLiteral("name")).toString());
+    if (attributes.value(QStringLiteral("type")).toString().compare(QLatin1String("destructor")) == 0) {
+        request.markAsDestructor();
+    }
     while (!m_xmlReader.atEnd()) {
         if (!m_xmlReader.readNextStartElement()) {
             if (m_xmlReader.qualifiedName().compare(QLatin1String("request")) == 0) {
@@ -429,6 +425,27 @@ void Generator::startGenerateServerHeaderFile()
         generateHeaderIncludes();
         generateStartNamespace();
         generateNamespaceForwardDeclarations();
+        if (std::any_of(m_interfaces.constBegin(), m_interfaces.constEnd(), [] (const Interface &i) { return i.isUnstableInterface(); })) {
+            // generate the unstable semantic version
+            auto it = std::find_if(m_interfaces.constBegin(), m_interfaces.constEnd(), [] (const Interface &i) { return i.isGlobal(); });
+            if (it != m_interfaces.constEnd()) {
+                const QString templateString = QStringLiteral(
+"/**\n"
+" * Enum describing the interface versions the %1 can support.\n"
+" *\n"
+" * @since 5.XX\n"
+" **/\n"
+"enum class %1Version {\n"
+"    /**\n"
+"     * %2\n"
+"     **/\n"
+"     UnstableV%3\n"
+"};\n\n");
+                *m_stream.localData() << templateString.arg((*it).kwaylandServerName())
+                                                       .arg((*it).name())
+                                                       .arg((*it).name().mid((*it).name().lastIndexOf(QStringLiteral("_v")) + 2));
+            }
+        }
         for (auto it = m_interfaces.constBegin(); it != m_interfaces.constEnd(); ++it) {
             generateClass(*it);
         }
@@ -597,7 +614,8 @@ void Generator::generateCppIncludes()
     switch (m_project.localData()) {
     case Project::Client:
         *m_stream.localData() << QStringLiteral("#include \"%1.h\"\n").arg(m_baseFileName.toLower());
-        *m_stream.localData() << QStringLiteral("#include \"event_queue.h\"\n\n");
+        *m_stream.localData() << QStringLiteral("#include \"event_queue.h\"\n");
+        *m_stream.localData() << QStringLiteral("#include \"wayland_pointer_p.h\"\n\n");
         break;
     case Project::Server:
         *m_stream.localData() << QStringLiteral("#include \"%1_interface.h\"\n").arg(m_baseFileName.toLower());
@@ -661,6 +679,10 @@ void Generator::generateClientResourceClass(const Interface &interface)
 
 void Generator::generateServerGlobalClass(const Interface &interface)
 {
+    if (interface.isUnstableInterface()) {
+        generateServerGlobalClassUnstable(interface);
+        return;
+    }
     const QString templateString = QStringLiteral(
 "class KWAYLANDSERVER_EXPORT %1 : public Global\n"
 "{\n"
@@ -677,8 +699,37 @@ void Generator::generateServerGlobalClass(const Interface &interface)
     *m_stream.localData() << templateString.arg(interface.kwaylandServerName());
 }
 
+void Generator::generateServerGlobalClassUnstable(const Interface &interface)
+{
+    const QString templateString = QStringLiteral(
+"class KWAYLANDSERVER_EXPORT %1 : public Global\n"
+"{\n"
+"    Q_OBJECT\n"
+"public:\n"
+"    virtual ~%1();\n"
+"\n"
+"    /**\n"
+"     * @returns The interface version used by this %1\n"
+"     **/\n"
+"    %1Version interfaceVersion() const;\n"
+"\n"
+"protected:\n"
+"    class Private;\n"
+"    explicit %1(Private *d, QObject *parent = nullptr);\n"
+"\n"
+"private:\n"
+"    Private *d_func() const;\n"
+"};\n"
+"\n");
+    *m_stream.localData() << templateString.arg(interface.kwaylandServerName());
+}
+
 void Generator::generateServerResourceClass(const Interface &interface)
 {
+    if (interface.factory()->isUnstableInterface()) {
+        generateServerResourceClassUnstable(interface);
+        return;
+    }
     const QString templateString = QStringLiteral(
 "class KWAYLANDSERVER_EXPORT %1 : public Resource\n"
 "{\n"
@@ -697,6 +748,32 @@ void Generator::generateServerResourceClass(const Interface &interface)
     *m_stream.localData() << templateString.arg(interface.kwaylandServerName()).arg(interface.factory()->kwaylandServerName());
 }
 
+void Generator::generateServerResourceClassUnstable(const Interface &interface)
+{
+    const QString templateString = QStringLiteral(
+"class KWAYLANDSERVER_EXPORT %1 : public Resource\n"
+"{\n"
+"    Q_OBJECT\n"
+"public:\n"
+"\n"
+"    virtual ~%1();\n"
+"\n"
+"    /**\n"
+"     * @returns The interface version used by this %1\n"
+"     **/\n"
+"    %2Version interfaceVersion() const;\n"
+"\n"
+"protected:\n"
+"    class Private;\n"
+"    explicit %1(Private *p, QObject *parent = nullptr);\n"
+"\n"
+"private:\n"
+"    Private *d_func() const;\n"
+"};\n"
+"\n");
+    *m_stream.localData() << templateString.arg(interface.kwaylandServerName()).arg(interface.factory()->kwaylandServerName());
+}
+
 void Generator::generatePrivateClass(const Interface &interface)
 {
     switch (m_project.localData()) {
@@ -752,6 +829,9 @@ void Generator::generateServerPrivateGlobalClass(const Interface &interface)
 void Generator::generateServerPrivateCallbackDefinitions(const Interface &interface)
 {
     for (const auto &r: interface.requests()) {
+        if (r.isDestructor() && !interface.isGlobal()) {
+            continue;
+        }
         *m_stream.localData() << QStringLiteral("    static void %1Callback(wl_client *client, wl_resource *resource").arg(toCamelCase(r.name()));
         for (const auto &a: r.arguments()) {
             *m_stream.localData() << QStringLiteral(", %1 %2").arg(a.typeAsServerWl()).arg(a.name());
@@ -764,14 +844,24 @@ void Generator::generateServerPrivateCallbackDefinitions(const Interface &interf
 void Generator::generateServerPrivateCallbackImpl(const Interface &interface)
 {
     for (const auto &r: interface.requests()) {
+        if (r.isDestructor() && !interface.isGlobal()) {
+            continue;
+        }
         *m_stream.localData() << QStringLiteral("void %2::Private::%1Callback(wl_client *client, wl_resource *resource").arg(toCamelCase(r.name())).arg(interface.kwaylandServerName());
         for (const auto &a: r.arguments()) {
             *m_stream.localData() << QStringLiteral(", %1 %2").arg(a.typeAsServerWl()).arg(a.name());
         }
         *m_stream.localData() << QStringLiteral(
 ")\n"
-"{\n"
-"    // TODO: implement\n"
+"{\n");
+        if (r.isDestructor()) {
+            *m_stream.localData() << QStringLiteral(
+"    Q_UNUSED(client)\n"
+"    wl_resource_destroy(resource);\n");
+        } else {
+            *m_stream.localData() << QStringLiteral("    // TODO: implement\n");
+        }
+        *m_stream.localData() << QStringLiteral(
 "}\n"
 "\n");
     }
@@ -847,7 +937,11 @@ void Generator::generateServerPrivateInterfaceClass(const Interface &interface)
         } else {
             first = false;
         }
-        *m_stream.localData() << QStringLiteral("    %1Callback").arg(toCamelCase(r.name()));
+        if (r.isDestructor() && !interface.isGlobal()) {
+            *m_stream.localData() << QStringLiteral("    resourceDestroyedCallback");
+        } else {
+            *m_stream.localData() << QStringLiteral("    %1Callback").arg(toCamelCase(r.name()));
+        }
     }
     *m_stream.localData() << QStringLiteral("\n};\n#endif\n\n");
 }
@@ -871,6 +965,55 @@ void Generator::generateServerPrivateResourceCtorDtorClass(const Interface &inte
 }
 
 void Generator::generateClientPrivateClass(const Interface &interface)
+{
+    if (interface.isGlobal()) {
+        generateClientPrivateGlobalClass(interface);
+    } else {
+        generateClientPrivateResourceClass(interface);
+    }
+
+    const auto events = interface.events();
+    if (!events.isEmpty()) {
+        *m_stream.localData() << QStringLiteral("\nprivate:\n");
+        // generate the callbacks
+        for (auto event : events) {
+            const QString templateString = QStringLiteral("    static void %1Callback(void *data, %2 *%2");
+            *m_stream.localData() << templateString.arg(event.name()).arg(interface.name());
+            const auto arguments = event.arguments();
+            for (auto argument : arguments) {
+                if (argument.interface().isNull()) {
+                    *m_stream.localData() << QStringLiteral(", %1 %2").arg(argument.typeAsServerWl()).arg(argument.name());
+                } else {
+                    *m_stream.localData() << QStringLiteral(", %1 *%2").arg(argument.interface()).arg(argument.name());
+                }
+            }
+            *m_stream.localData() << ");\n";
+        }
+        *m_stream.localData() << QStringLiteral("\n    static const %1_listener s_listener;\n").arg(interface.name());
+    }
+
+    *m_stream.localData() << QStringLiteral("};\n\n");
+}
+
+void Generator::generateClientPrivateResourceClass(const Interface &interface)
+{
+    const QString templateString = QStringLiteral(
+"class %1::Private\n"
+"{\n"
+"public:\n"
+"    Private(%1 *q);\n"
+"\n"
+"    void setup(%2 *arg);\n"
+"\n"
+"    WaylandPointer<%2, %2_destroy> %3;\n"
+"\n"
+"private:\n"
+"    %1 *q;\n");
+
+    *m_stream.localData() << templateString.arg(interface.kwaylandClientName()).arg(interface.name()).arg(interface.kwaylandClientName().toLower());
+}
+
+void Generator::generateClientPrivateGlobalClass(const Interface &interface)
 {
     const QString templateString = QStringLiteral(
 "class %1::Private\n"
@@ -878,24 +1021,101 @@ void Generator::generateClientPrivateClass(const Interface &interface)
 "public:\n"
 "    Private() = default;\n"
 "\n"
-"    WaylandPointer<%2, %2_destroy> %3;\n");
+"    void setup(%2 *arg);\n"
+"\n"
+"    WaylandPointer<%2, %2_destroy> %3;\n"
+"    EventQueue *queue = nullptr;\n");
 
     *m_stream.localData() << templateString.arg(interface.kwaylandClientName()).arg(interface.name()).arg(interface.kwaylandClientName().toLower());
-    if (interface.isGlobal()) {
-        *m_stream.localData() << "    EventQueue *queue = nullptr;\n";
-    }
-    *m_stream.localData() << "};\n\n";
 }
 
 void Generator::generateClientCpp(const Interface &interface)
 {
-    const QString templateString = QStringLiteral(
+    // TODO: generate listener and callbacks
+    const auto events = interface.events();
+    if (!events.isEmpty()) {
+        // listener
+        *m_stream.localData() << QStringLiteral("const %1_listener %2::Private::s_listener = {\n").arg(interface.name()).arg(interface.kwaylandClientName());
+        bool first = true;
+        for (auto event : events) {
+            if (!first) {
+                *m_stream.localData() << QStringLiteral(",\n");
+            }
+            *m_stream.localData() << QStringLiteral("    %1Callback").arg(event.name());
+            first = false;
+        }
+        *m_stream.localData() << QStringLiteral("\n};\n\n");
+
+        // callbacks
+        for (auto event : events) {
+            *m_stream.localData() << QStringLiteral("void %1::Private::%2Callback(void *data, %3 *%3").arg(interface.kwaylandClientName()).arg(event.name()).arg(interface.name());
+
+            const auto arguments = event.arguments();
+            for (auto argument : arguments) {
+                if (argument.interface().isNull()) {
+                    *m_stream.localData() << QStringLiteral(", %1 %2").arg(argument.typeAsServerWl()).arg(argument.name());
+                } else {
+                    *m_stream.localData() << QStringLiteral(", %1 *%2").arg(argument.interface()).arg(argument.name());
+                }
+            }
+
+            *m_stream.localData() << QStringLiteral(
+")\n"
+"{\n"
+"    auto p = reinterpret_cast<%1::Private*>(data);\n"
+"    Q_ASSERT(p->%2 == %3);\n").arg(interface.kwaylandClientName())
+                               .arg(interface.kwaylandClientName().toLower())
+                               .arg(interface.name());
+            for (auto argument : arguments) {
+                *m_stream.localData() << QStringLiteral("    Q_UNUSED(%1)\n").arg(argument.name());
+            }
+            *m_stream.localData() << QStringLiteral("    // TODO: implement\n}\n\n");
+        }
+    }
+
+    if (interface.isGlobal()) {
+        // generate ctor without this pointer to Private
+        const QString templateString = QStringLiteral(
 "%1::%1(QObject *parent)\n"
 "    : QObject(parent)\n"
 "    , d(new Private)\n"
 "{\n"
+"}\n");
+        *m_stream.localData() << templateString.arg(interface.kwaylandClientName());
+    } else {
+        // Private ctor
+        const QString templateString = QStringLiteral(
+"%1::Private::Private(%1 *q)\n"
+"    : q(q)\n"
+"{\n"
 "}\n"
 "\n"
+"%1::%1(QObject *parent)\n"
+"    : QObject(parent)\n"
+"    , d(new Private(this))\n"
+"{\n"
+"}\n");
+        *m_stream.localData() << templateString.arg(interface.kwaylandClientName());
+    }
+
+    // setup call with optional add_listener
+    const QString setupTemplate = QStringLiteral(
+"\n"
+"void %1::Private::setup(%3 *arg)\n"
+"{\n"
+"    Q_ASSERT(arg);\n"
+"    Q_ASSERT(!%2);\n"
+"    %2.setup(arg);\n");
+    *m_stream.localData() << setupTemplate.arg(interface.kwaylandClientName())
+                                           .arg(interface.kwaylandClientName().toLower())
+                                           .arg(interface.name());
+    if (!interface.events().isEmpty()) {
+        *m_stream.localData() << QStringLiteral("    %1_add_listener(%2, &s_listener, this);\n").arg(interface.name()).arg(interface.kwaylandClientName().toLower());
+    }
+    *m_stream.localData() << QStringLiteral("}\n");
+
+    const QString templateString = QStringLiteral(
+"\n"
 "%1::~%1()\n"
 "{\n"
 "    release();\n"
@@ -903,9 +1123,7 @@ void Generator::generateClientCpp(const Interface &interface)
 "\n"
 "void %1::setup(%3 *%2)\n"
 "{\n"
-"    Q_ASSERT(%2);\n"
-"    Q_ASSERT(!d->%2);\n"
-"    d->%2.setup(%2);\n"
+"    d->setup(%2);\n"
 "}\n"
 "\n"
 "void %1::release()\n"
@@ -918,16 +1136,6 @@ void Generator::generateClientCpp(const Interface &interface)
 "    d->%2.destroy();\n"
 "}\n"
 "\n"
-"void %1::setEventQueue(EventQueue *queue)\n"
-"{\n"
-"    d->queue = queue;\n"
-"}\n"
-"\n"
-"EventQueue *%1::eventQueue()\n"
-"{\n"
-"    return d->queue;\n"
-"}\n"
-"\n"
 "%1::operator %3*() {\n"
 "    return d->%2;\n"
 "}\n"
@@ -944,6 +1152,20 @@ void Generator::generateClientCpp(const Interface &interface)
     *m_stream.localData() << templateString.arg(interface.kwaylandClientName())
                                            .arg(interface.kwaylandClientName().toLower())
                                            .arg(interface.name());
+    if (interface.isGlobal()) {
+        const QString templateStringGlobal = QStringLiteral(
+"void %1::setEventQueue(EventQueue *queue)\n"
+"{\n"
+"    d->queue = queue;\n"
+"}\n"
+"\n"
+"EventQueue *%1::eventQueue()\n"
+"{\n"
+"    return d->queue;\n"
+"}\n\n"
+        );
+        *m_stream.localData() << templateStringGlobal.arg(interface.kwaylandClientName());
+    }
 }
 
 void Generator::generateClientGlobalClassDoxy(const Interface &interface)
@@ -1122,13 +1344,31 @@ void Generator::generateClientClassRequests(const Interface &interface)
 void Generator::generateClientCppRequests(const Interface &interface)
 {
     const auto requests = interface.requests();
-    const QString templateString = QStringLiteral("void %1::%2(%3)\n");
-    const QString factoryTemplateString = QStringLiteral("%2 *%1::%3(%4)\n");
+    const QString templateString = QStringLiteral(
+"void %1::%2(%3)\n"
+"{\n"
+"    Q_ASSERT(isValid());\n"
+"    %4_%5(d->%6%7);\n"
+"}\n\n");
+    const QString factoryTemplateString = QStringLiteral(
+"%2 *%1::%3(%4)\n"
+"{\n"
+"    Q_ASSERT(isValid());\n"
+"    auto p = new %2(parent);\n"
+"    auto w = %5_%6(d->%7%8);\n"
+"    if (d->queue) {\n"
+"        d->queue->addProxy(w);\n"
+"    }\n"
+"    p->setup(w);\n"
+"    return p;\n"
+"}\n\n"
+    );
     for (const auto &r: requests) {
         if (r.isDestructor()) {
             continue;
         }
         QString arguments;
+        QString requestArguments;
         bool first = true;
         QString factored;
         for (const auto &a: r.arguments()) {
@@ -1143,20 +1383,38 @@ void Generator::generateClientCppRequests(const Interface &interface)
             }
             if (a.type() == Argument::Type::Object) {
                 arguments.append(QStringLiteral("%1 *%2").arg(a.typeAsQt()).arg(toCamelCase(a.name())));
+                requestArguments.append(QStringLiteral(", *%1").arg(toCamelCase(a.name())));
             } else {
                 arguments.append(QStringLiteral("%1 %2").arg(a.typeAsQt()).arg(toCamelCase(a.name())));
+                QString arg = toCamelCase(a.name());
+                if (a.type() == Argument::Type::Fixed) {
+                    arg = QStringLiteral("wl_fixed_from_double(%1)").arg(arg);
+                }
+                requestArguments.append(QStringLiteral(", %1").arg(arg));
             }
         }
         if (factored.isEmpty()) {
-            *m_stream.localData() << templateString.arg(interface.kwaylandClientName()).arg(toCamelCase(r.name())).arg(arguments);
+            *m_stream.localData() << templateString.arg(interface.kwaylandClientName())
+                                                   .arg(toCamelCase(r.name()))
+                                                   .arg(arguments)
+                                                   .arg(interface.name())
+                                                   .arg(r.name())
+                                                   .arg(interface.kwaylandClientName().toLower())
+                                                   .arg(requestArguments);
         } else {
             if (!first) {
                 arguments.append(QStringLiteral(", "));
             }
             arguments.append(QStringLiteral("QObject *parent"));
-            *m_stream.localData() << factoryTemplateString.arg(interface.kwaylandClientName()).arg(toQtInterfaceName(factored)).arg(toCamelCase(r.name())).arg(arguments);
+            *m_stream.localData() << factoryTemplateString.arg(interface.kwaylandClientName())
+                                                          .arg(toQtInterfaceName(factored))
+                                                          .arg(toCamelCase(r.name()))
+                                                          .arg(arguments)
+                                                          .arg(interface.name())
+                                                          .arg(r.name())
+                                                          .arg(interface.kwaylandClientName().toLower())
+                                                          .arg(requestArguments);
         }
-        *m_stream.localData() << QStringLiteral("{\n}\n\n");
     }
 }
 
